(**
   This is a module on confluence.
@author Naoki Nishida
@since 2014/06/26
@since 2016/11/20
*)

open Debug;;
open Ctrs;;

(**
   This decides whether a given CTRS is orthogonal.
*)
let is_orthogonal spec =
  is_LL spec && not (Cpair.is_overlapping spec.rules);;

(**
   This checks whether a given CTRS is confluent.
*)
let is_CR spec = 
  if is_orthogonal spec 
  then
    begin
      debug ("Succeeded in proving orthogonality of the following CTRS:\n"^(string_of_ctrs 2 spec));
      true
    end
  else
    if Sn.is_SN spec = Some true
    then
      begin
	debug ("Succeeded in proving termination of the following CTRS:\n"^(string_of_ctrs 2 spec));
	if List.for_all (Cpair.is_joinable spec) (Cpair.all_cpairs spec)
	then
	  begin
	    debug "Succeeded in proving joinablity of all CPs.\n";
	    true
	  end
	else
	  begin
	    debug "Exists a CP that is not joinable.\n";
	    debug "Failed to prove confluence of the above CTRS.\n";
	    false
	  end
      end
    else
      begin
	debug ("Failed to prove confluence of the following CTRS:\n"^(string_of_ctrs 2 spec));
	false
      end;;

(**
   This checks whether a given CTRS is non-confluent.
*)
let is_not_CR spec =
  let spec_u = unconditional spec in
  let cps = Cpair.all_cpairs spec_u in
  let unjoinable (s,t) =
    begin
      let (s1,t1) = Dp.cap spec (s,t) in
      begin
	try
	  let _ = unify s1 t1 in
	  false
	with
	  Not_found ->true
      end
    end
  ||
    (is_ground_strongly_irreducible_term spec s
      && is_ground_strongly_irreducible_term spec t
      && s <> t)
  in
(*
  print_string "(COMMENT unconditional CPs are\n";
  let _ = List.map (fun (s,t) ->
    print_string ("  ("^(string_of_term s)^", "^(string_of_term t)^")\n")) cps
  in
  print_string ")\n";
*)
  List.exists unjoinable cps;;
