(**
    This is a module to deal with extended narrowing trees.

    @since 2017/04/28
    @author Naoki Nishida
 *)

open Util;;
open Variable;;
open Ctrs;;

(**
    This is a data type for equations and their sets.
 *)
type eqn = term * term;;
type eqnset = eqn list;;

(**
   This is a data type for substitution-set expressions.
*)
type substsetexpr =
  EmptySet
| Single of subst
| SComposition of substsetexpr * substsetexpr
| PComposition of substsetexpr * vid list * substsetexpr
| MGU of eqnexpr
| REC of renaming * nonterminal
and eqnexpr =
  EqnSet of eqnset
| Apply of eqnexpr * substsetexpr
| EQN of substsetexpr
and nonterminal =
  Gamma
;;

(**
 This simplifies [substsetexpr].
*)
let rec simplify_substsetexpr e =
begin
    match e with
      SComposition (e1,e2) ->
        begin
          let (e3,e4) = (simplify_substsetexpr e1, simplify_substsetexpr e2) in
          match (e3,e4) with
            (_,EmptySet)
          | (EmptySet,_) -> EmptySet
          | _ -> SComposition (e3,e4)
        end
    | PComposition (e1,vs,e2) ->
        begin
          let (e3,e4) = (simplify_substsetexpr e1, simplify_substsetexpr e2) in
          match (e3,e4) with
            (EmptySet,_)
          | (_,EmptySet) -> EmptySet
          | _ -> PComposition (e3,vs,e4)
        end
    | _ -> e 
end;;
