(**
   This is a module to rewrite terms.
   @author Naoki Nishida
   @since 2014/06/27
   @since 2016/11/22
*)
  
open Ctrs;;
  
(**
   This computes all reducts of a given term.
*)
let rec all_reducts rs =
  function
      Node(Var _,_) -> []
    | (Node(FSym f,ts) as t) ->
	let rec all_reducts_at_root =
	  function
	      [] -> []
	    | (Rew (l,r,[]))::rs1 ->
		begin
		  try
		    let theta = match_term l t in
		    (apply_subst_to_term theta r)::(all_reducts_at_root rs1)
		  with
		    Not_matching ->
		      all_reducts_at_root rs1
		end
	    | _ -> raise Unsupported
	and all_reducts_of_terms =
	  function
	      [] -> []
	    | t::ts ->
		(List.map (fun u -> u::ts) (all_reducts rs t))
		@ (List.map (fun us -> t::us) (all_reducts_of_terms ts))
	in
	(all_reducts_at_root rs)
	@ (List.map (fun us -> Node(FSym f,us)) (all_reducts_of_terms ts));;
      
(**
   This computes all normal forms of a given term.
*)
let all_nfs rs t = 
  let nfs = ref [] in
  let rec get_nfs =
    function
	[] -> !nfs
      |	(t::ts) ->
	  begin
	    match all_reducts rs t with
	      [] ->
		begin
		  nfs := t::(!nfs);
		  get_nfs ts
		end
	    | us -> get_nfs (us@ts)
	  end
  in
  get_nfs [t];;

(**
   This rewrites are redex by the first applicable rule.
*)
let rec rewrite rs t =
  match rs with 
    [] -> None
  | (Rew (l,r,[]))::rs1 ->
      begin
	try
	  let theta = match_term l t in
	  Some (apply_subst_to_term theta r)
	with
	  Not_matching ->
	    rewrite rs1 t
      end
  | _ -> raise Unsupported;;


(**
   This computes a normal form by leftmost innermost strategy.
*)
let rec linf rs t =
(*  match t with
    Node(Var _,_) -> t
  | Node(f,ts) -> 
      begin
	let t1 = Node(f,List.map (linf rs) ts) in
	match rewrite rs t1 with
	  None -> t1
	| Some t2 -> linf rs t2
      end;;
*)
  match t with
    Node(Var _,_) -> t
  | Node(f,ts) ->
      let rec linftop rs0 t0 =
	match rs0 with
	  [] -> t0
	| (Rew (l,r,[]))::rs1 ->
	    begin
	      try
		let theta = match_term l t0 in
		linfsubst theta r
	      with
		Not_matching -> linftop rs1 t0
	    end
	| _ -> raise Unsupported 
      and linfsubst theta t1 =
	match t1 with
	  Node(Var x,_) ->
	    begin
	      try
		let u = List.assoc x theta in
		u
	      with
		Not_found -> t1
	    end
	| Node(f,ts) ->
	    let t2 = Node(f,List.map (linfsubst theta) ts) in
	    linftop rs t2
      in
      linftop rs (Node(f,List.map (linf rs) ts));;
