(**
   This provides utility operations.
   @author Naoki Nishida
   @since 2010/12/17
*)

(** 
    This explodes strings to character lists,
    e.g.,
    [explode "abc" = \['a';'b';'c'\]], 
    [explode "" = \[\]]
*)
(*
let explode str =
  let res = ref []
  in
    for i = 0 to (String.length str)-1 do
      res := !res @ [str.[i]];
    done;
    !res;;
*)
let explode s =
  let rec expl i l =
    if i < 0 then l else
    expl (i - 1) (s.[i] :: l) in
  expl (String.length s - 1) [];;


(**
   This converts characters to strings, 
   e.g., [string_of_char 'a' = "a"]
*)
let string_of_char a =
  let s = "a"
  in
    String.set s 0 a; (* s.[0] <- a; *)
    s;;

(** 
    This implodes character lists to strings,
    e.g., [implode \['a';'b';'c'\] = "abc"], [implode \[\] = ""]
*)
(*
let rec implode cs =
  match cs with
      [] -> ""
    | c::cs1 -> (string_of_char c)^(implode cs1);;
*)
let implode l =
  let result = String.create (List.length l) in
  let rec imp i = function
  | [] -> result
  | c :: l ->
      begin
	String.set result i c; (* result.[i] <- c; *)
	imp (i + 1) l
      end
  in
  imp 0 l;;

(**
   This is a data type for sets.
*)
type 'a set = 'a list;;

(**
   This is the union operation of lists.
*)
let rec union xs =
  function
      [] -> xs
    | y::ys -> if List.mem y xs then union xs ys else union (xs@[y]) ys;;

(**
   This is the setminus operation of lists.
*)
let rec setminus xs0 ys =
  begin
    match xs0 with
      [] -> []
    | x::xs -> if List.mem x ys then setminus xs ys else x::(setminus xs ys)
  end;;

(**
   This is the intersection operation of lists.
*)
let rec intersection xs =
  function
      [] -> []
    | y::ys -> 
	if List.mem y xs then y::(intersection xs ys)
	else intersection xs ys;;

(**
   This returns true if the intersection of the two input lists is empty.
*)
let rec is_intersectionEmpty xs =
  function
      [] -> true
    | y::ys -> (not (List.mem y xs)) && is_intersectionEmpty xs ys;;

(**
   This is an exception for the failure of appending association lists.
*)
exception Not_Append;;

(**
   This is the append operation of association lists.
*)
let rec assoc_append xs =
  function
      [] -> xs
    | (y,v)::ys ->
	try
	  let v1 = List.assoc y xs in
	  if v = v1 then assoc_append xs ys else raise Not_Append
	with
	  Not_found -> (y,v)::(assoc_append xs ys);;

(**
   This is a basic indent.
*)
let indent = "  ";;
    
(**
   This generates [n] indents.
*)
let rec string_of_indent n =
  if n <= 0 then "" else (indent)^(string_of_indent (n-1));;

(**
   This decides whether the 1st set is a subset of the 2nd set.
*)
let rec is_subseteq xs ys =
    match xs with
      [] -> true
    | x::xs -> List.mem x ys && is_subseteq xs ys;;

(**
   This returns the first [n] elements of a given list.
*)
let rec take n =
  function
      [] -> []
    | (x::xs) -> if n > 0 then x::(take (n-1) xs) else [];;
  
